""" functions to load genes, and identify transcripts containing de novos.
"""

import asyncio

from denovonear.transcript import Transcript

from denovonear.ensembl_requester import (get_protein_seq_for_transcript,
    get_genomic_seq_for_transcript, get_cds_seq_for_transcript,
    get_cds_ranges_for_transcript, get_exon_ranges_for_transcript,
    get_genes_for_hgnc_id, get_transcript_ids_for_ensembl_gene_id,
    get_previous_symbol)

async def get_transcript_lengths(ensembl, transcript_ids):
    """ finds the protein length for ensembl transcript IDs for a gene
    
    Args:
        ensembl: EnsemblRequest object to request sequences and data
            from the ensembl REST API
        transcript_ids: list of transcript IDs for a single gene
    
    Returns:
        dictionary of lengths (in amino acids), indexed by transcript IDs
    """
    tasks = [get_protein_seq_for_transcript(ensembl, x) for x in transcript_ids]
    seqs = await asyncio.gather(*tasks, return_exceptions=True)
    
    # associate tx_ids to the protein sequences, but remove failures
    transcripts = dict(zip(transcript_ids, seqs))
    for key in list(transcripts.keys()):
        if not isinstance(transcripts[key], str):
            del transcripts[key]
        else:
            transcripts[key] = len(transcripts[key])
    
    return transcripts

async def construct_gene_object(ensembl, transcript_id):
    """ creates an Transcript object for a gene from ensembl databases
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        transcript_id: string for an Ensembl transcript ID
    
    Returns:
        a Transcript object, containing transcript coordinates and gene and
        transcript sequence.
    
    Raises:
        ValueError if CDS from genomic sequence given gene coordinates and CDS
        retrieved from Ensembl do not match.
    """
    tasks = [get_genomic_seq_for_transcript(ensembl, transcript_id, expand=10),
        get_cds_seq_for_transcript(ensembl, transcript_id),
        get_cds_ranges_for_transcript(ensembl, transcript_id),
        get_exon_ranges_for_transcript(ensembl, transcript_id)]
    (chrom, start, end, strand, genomic), cds_seq, cds, exons = await asyncio.gather(*tasks)
    
    # start a Transcript object with the locations and sequence
    transcript = Transcript(transcript_id, chrom, start, end, strand)
    transcript.set_exons(exons, cds)
    transcript.set_cds(cds)
    transcript.add_cds_sequence(cds_seq)
    transcript.add_genomic_sequence(genomic, offset=10)
    
    return transcript

def get_de_novos_in_transcript(transcript, de_novos):
    """ get the de novos within the coding sequence of a transcript
    
    Args:
        transcript: Transcript object, which defines the transcript coordinates
        de_novos: list of chromosome sequence positions for de novo events
    
    Returns:
        list of de novo positions found within the transcript
    """
    
    in_transcript = []
    for de_novo in de_novos:
        # we check if the de novo is within the transcript by converting the
        # chromosomal position to a CDS-based position. Variants outside the CDS
        # will raise an error, which we catch and pass on. It's better to do
        # this, rather than use the function in_coding_region(), since that
        # function does not allow for splice site variants.
        site = transcript.get_coding_distance(de_novo)
        cds_length = transcript.get_coding_distance(transcript.get_cds_end())
        within_cds = site['pos'] >= 0 and site['pos'] < cds_length['pos']
        if within_cds and (transcript.in_coding_region(de_novo) or abs(site['offset']) < 9):
            in_transcript.append(de_novo)
    
    return in_transcript

def flatten(values):
    return [x for sub in values for x in sub]

async def get_transcript_ids(ensembl, symbol):
    """ gets transcript IDs for a gene.
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        symbol: HGNC symbol for gene
    
    Returns:
        dictionary of transcript ID: transcript lengths for all transcripts
        for a given HGNC symbol.
    """
    
    genes = await get_genes_for_hgnc_id(ensembl, symbol)
    tasks = [get_transcript_ids_for_ensembl_gene_id(ensembl, x, symbol) for x in genes]
    transcript_ids = flatten(await asyncio.gather(*tasks))
    
    # sometimes we get HGNC symbols that do not match the ensembl rest version
    # that we are currentl using. We can look for earlier HGNC symbols for
    # the gene using the service at rest.genenames.org
    symbols = []
    if len(transcript_ids) == 0:
        symbols = await get_previous_symbol(ensembl, symbol)
        tasks = [get_genes_for_hgnc_id(ensembl, s) for s in symbols]
        genes += flatten(await asyncio.gather(*tasks))
        symbols = [symbol] + symbols
        
        tasks = [get_transcript_ids_for_ensembl_gene_id(ensembl, x, symbols) for x in genes]
        transcript_ids = flatten(await asyncio.gather(*tasks))
    
    return await get_transcript_lengths(ensembl, transcript_ids)

async def load_gene(ensembl, gene_id, de_novos=[]):
    """ sort out all the necessary sequences and positions for a gene
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        gene_id: HGNC symbol for gene
        de_novos: list of de novo positions, so we can check they all fit in
            the gene transcript
        
    Returns:
        list of Transcript objects for gene, including genomic ranges and sequences
    """
    
    transcripts = await minimise_transcripts(ensembl, gene_id, de_novos)
    
    genes = []
    for transcript_id in transcripts:
        gene = await construct_gene_object(ensembl, transcript_id)
        genes.append(gene)
    
    if len(genes) == 0:
        raise IndexError("{0}: no suitable transcripts".format(gene_id))
    
    return genes
    
async def count_de_novos_per_transcript(ensembl, gene_id, de_novos=[]):
    """ count de novos in transcripts for a gene.
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        gene_id: HGNC symbol for gene
        de_novos: list of de novo positions, so we can check they all fit in
            the gene transcript
        
    Returns:
        dictionary of lengths and de novo counts, indexed by transcript IDs.
    """
    
    transcripts = await get_transcript_ids(ensembl, gene_id)
    
    # TODO: allow for genes without any coding sequence.
    if len(transcripts) == 0:
        raise IndexError("{0} lacks coding transcripts".format(gene_id))
    
    # count the de novos observed in all transcripts
    counts = {}
    for key in transcripts:
        try:
            gene = await construct_gene_object(ensembl, key)
            total = len(get_de_novos_in_transcript(gene, de_novos))
            if total > 0:
                counts[key] = {}
                counts[key]["n"] = total
                counts[key]["len"] = transcripts[key]
        except ValueError:
            pass
    
    return counts

async def minimise_transcripts(ensembl, gene_id, de_novos):
    """ get a set of minimal transcripts to contain all the de novos.
    
    We identify the minimal number of transcripts to contain all de novos. This
    allows for de novos on mutually exclusive transcripts. The transcripts are
    selected on the basis of containing the most number of de novos, while also
    being the longest possible transcript for the gene.
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        gene_id: HGNC symbol for gene
        de_novos: set of de novo positions
    
    Returns:
        dictionary of lengths and de novo counts, indexed by transcript ID for
        the set of minimal transcripts necessary to contain all de novos.
    """
    
    if len(de_novos) == 0:
        return {}
    
    counts = await count_de_novos_per_transcript(ensembl, gene_id, de_novos)
    
    if len(counts) == 0:
        return {}
    
    # find the transcripts with the most de novos
    max_count = max( val["n"] for key, val in counts.items() )
    transcripts = [ key for key, val in counts.items() if val["n"] == max_count ]
    
    # find the transcripts with the greatest length, and the most de novos
    max_length = max( counts[key]["len"] for key in transcripts )
    tx_ids = [ key for key in transcripts if counts[key]["len"] == max_length ]
    max_transcripts = {x: counts[x] for x in counts if x in tx_ids}
    
    # find which de novos occur in the transcript with the most de novos
    gene = await construct_gene_object(ensembl, next(iter(max_transcripts)))
    denovos_in_gene = get_de_novos_in_transcript(gene, de_novos)
    
    # trim the de novos to the ones not in the current transcript
    leftovers = [ x for x in de_novos if x not in denovos_in_gene ]
    
    # and recursively return the transcripts in the current transcript, along
    # with transcripts for the reminaing de novos
    update = await minimise_transcripts(ensembl, gene_id, leftovers)
    max_transcripts.update(update)
    
    return max_transcripts



""" class to analyse clustering of known de novos in genes according to their
distances apart within the gene, and compare that to simulated de novo events
within the same gene.
"""

from denovonear.weights import geomean, get_distances, analyse_de_novos

def get_p_value(transcript, rates, iterations, consequence, de_novos):
    """ find the probability of getting de novos with a mean conservation
    
    The probability is the number of simulations where the mean conservation
    between simulated de novos is less than the observed conservation.
    
    Args:
        transcript: Transcript object for the current gene.
        rates: SiteRates object, which contains WeightedChoice entries for
            different consequence categories.
        iterations: number of simulations to perform
        consequence: string to indicate the consequence type e.g. "missense, or
            "lof", "synonymous" etc. The full list is "missense", "nonsense",
            "synonymous", "lof", "loss_of_function", "splice_lof",
            "splice_region".
        de_novos: list of de novos within a gene
    
    Returns:
        tuple of mean proximity for the observed de novos and probability of
        obtaining a value less than or equal to the observed proximity from the
        null distribution.
    """
    
    if len(de_novos) < 2:
        return (float('nan'), float('nan'))
    
    rename = {"lof": "loss_of_function"}
    if consequence in rename:
        consequence = rename[consequence]
    
    weights = rates[consequence]
    
    cds_positions = [ transcript.get_coding_distance(x)['pos'] for x in de_novos ]
    distances = get_distances(cds_positions)
    observed = geomean(distances)
    
    # call a cython wrapped C++ library to handle the simulations
    sim_prob = analyse_de_novos(weights, iterations, len(de_novos), observed)
    
    observed = "{0:0.1f}".format(observed)
    
    return (observed, sim_prob)




from pathlib import Path
import asyncio
import random
import functools

import aiohttp

from denovonear.ensembl_cache import EnsemblCache

def ensembl_cache(folder=None):
    ''' store/retrive repeated ensembl requests from a persistent sqlite cache
    '''
    if folder is None:
        folder = Path.home() / '.cache' / 'ensembl'
    cache = EnsemblCache(str(folder))
    def decorator(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            url = args[1]
            if 'rest.ensembl.org' in url:
                cached = cache.get_cached_data(url)
                if cached is not None:
                    return cached
                data = await func(*args, **kwargs)
                cache.cache_url_data(url, data)
                return data
            else:
                return await func(*args, **kwargs)
        return wrapper
    return decorator

def ensembl_retry(retries=5):
    ''' perform all the error handling for the request
    
    retries up to N times under certain error conditions, and increases waiting
    time between requests, unless we've hit rate limits, when it uses the stated
    retry time.
    '''
    def decorator(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            result = None
            last_exception = None
            for i in range(retries):
                try:
                    return await func(*args, **kwargs)
                except (aiohttp.ServerDisconnectedError, aiohttp.ClientOSError,
                        asyncio.TimeoutError) as err:
                    last_exception = err
                    delay = 0
                except aiohttp.ClientResponseError as err:
                    last_exception = err
                    # 500, 503, 504 are server down issues. 429 exceeds rate
                    # limits. 400 is server memory issue. Raises other errors.
                    if err.status not in [500, 503, 504, 429, 400]:
                        raise err
                    if err.status == 400:
                        if 'Cannot allocate memory' not in err.message:
                            raise err
                    delay = random.uniform(0, 2 ** (i + 2))
                    if err.status == 429:
                        delay = float(dict(err.headers)['Retry-After'])
                await asyncio.sleep(delay)
            if last_exception:
                raise last_exception
            return result
        return wrapper
    return decorator




from math import log, isnan

from scipy.stats import chi2

from denovonear.load_gene import load_gene, get_de_novos_in_transcript
from denovonear.load_mutation_rates import load_mutation_rates
from denovonear.load_de_novos import load_de_novos
from denovonear.site_specific_rates import SiteRates
from denovonear.simulate import get_p_value

def fishers_method(values):
    """ function to combine p values, using Fisher's method
    
    We occasionally have multiple P values for a mutation type, obtained from
    alternative transcripts for the gene. If we have only one P value for the
    gene for the mutation type, we just use that value, if we don't have any
    data, we use "NA", otherwise combine p-values using Fisher's method.
    
    Args:
        x: list of P-values for a gene
    
    Returns:
        combined P-value
    """
    values = [ x for x in values if not isnan(x) ]
    # use Fisher's combined method to estimate the P value from multiple
    # P-values. The chi square statistic is -2*sum(ln(P-values))
    return chi2.sf(-2 * sum(map(log, values)), 2 * len(values))

async def cluster_de_novos(symbol, de_novos, ensembl, iterations=1000000, mut_dict=None):
    """ analysis proximity cluster of de novos in a single gene
    
    Args:
        symbol: HGNC symbol for a gene
        de_novos: dictionary of de novo positions for the HGNC gene,
        indexed by functional type
        iterations: number of simulations to run
        ensembl: EnsemblRequest object, for obtaing info from ensembl
        mut_dict: dictionary of mutation rates, indexed by trinuclotide sequence
    
    Returns:
        a dictionary containing P values, and distances for missense, nonsense,
        and synonymous de novos events. Missing data is represented by "NA".
    """
    
    if mut_dict is None:
        mut_dict = load_mutation_rates()
    
    missense = de_novos["missense"]
    nonsense = de_novos["nonsense"]
    
    # load the set of transcripts that are the  minimum set of transcripts
    # required to contain all the de novos, unless we can't find any coding
    # transcripts that contain the de novos.
    try:
        transcripts = await load_gene(ensembl, symbol, missense + nonsense)
    except IndexError as e:
        print(e)
        return None
    
    probs = {"miss_prob": [], "nons_prob": []}
    dists = {"miss_dist": [], "nons_dist": []}
    
    for transcript in transcripts:
        
        missense_events = get_de_novos_in_transcript(transcript, missense)
        nonsense_events = get_de_novos_in_transcript(transcript, nonsense)
        
        rates = SiteRates(transcript, mut_dict)
        
        (miss_dist, miss_prob) = get_p_value(transcript, rates, iterations, "missense", missense_events)
        (nons_dist, nons_prob) = get_p_value(transcript, rates, iterations, "lof", nonsense_events)
        
        dists["miss_dist"].append(miss_dist)
        dists["nons_dist"].append(nons_dist)
        probs["miss_prob"].append(miss_prob)
        probs["nons_prob"].append(nons_prob)
        
        # remove the de novos analysed in the current transcript, so that
        # analysis of subsequent transcripts uses independent events. NOTE THAT
        # THIS MIGHT MISS SOME CLUSTERING ACROSS MUTUALLY EXCLUSIVE TRANSCRIPTS
        # IF THE DE NOVO EVENTS ARE NEAR THE TRANSCRIPT DIVERGENCE.
        missense = [x for x in missense if x not in missense_events]
        nonsense = [x for x in nonsense if x not in  nonsense_events]
        
    for key in dists:
        dists[key] = ",".join([ str(x) for x in dists[key] ])
    
    probs = {k: fishers_method(probs[k]) for k in probs}
    probs.update(dists)
    
    return probs




import time
import logging
import asyncio

import aiohttp

from denovonear.rate_limiter_retries import ensembl_retry as retry
from denovonear.rate_limiter_retries import ensembl_cache as cache

class RateLimiter:
    ''' class to asynchronously perform http get requests

    This respects the rate limits imposed by the server. Error handling and
    retrying are handled by the 'retry' decorator.
    '''
    MAX_TOKENS = 10
    def __init__(self, per_second=10):
        ''' initialize the class object

        Args:
            per_second: number of queries allowed per second
        '''
        self.tokens = self.MAX_TOKENS
        self.RATE = per_second
        self.updated_at = time.monotonic()
    async def __aenter__(self):
        self.client = aiohttp.ClientSession(trust_env=True)
        return self
    async def __aexit__(self, *err):
        await self.client.close()
        self.client = None

    @cache()
    @retry(retries=9)
    async def get(self, url, headers=None):
        ''' perform asynchronous http get

        Args:
            url: url to get
            headers: http headers to pass in with the get query
        '''
        if not headers:
            headers = {'content-type': 'application/json'}
        await self.wait_for_token()
        async with self.client.get(url, headers=headers) as resp:
            logging.info(f'{url}\t{resp.status}')
            resp.raise_for_status()
            return await resp.read()

    async def wait_for_token(self):
        ''' pause until tokens are refilled
        '''
        while self.tokens < 1:
            self.add_new_tokens()
            await asyncio.sleep(1 / self.RATE)
        self.tokens -= 1

    def add_new_tokens(self):
        ''' add new tokens if sufficient time has passed
        '''
        now = time.monotonic()
        time_since_update = now - self.updated_at
        new_tokens = time_since_update * self.RATE
        if self.tokens + new_tokens >= 1:
            self.tokens = min(self.tokens + new_tokens, self.MAX_TOKENS)
            self.updated_at = now



# cython: language_level=3, boundscheck=False
""" get weight gene mutation rates
"""

from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp cimport bool
from cython.operator cimport dereference as deref

from denovonear.weights cimport Chooser, WeightedChoice
from denovonear.transcript cimport Tx, Transcript, Region

cdef extern from "site_rates.h":
    cdef cppclass SitesChecks:
        SitesChecks(Tx, vector[vector[string]], bool) except +
        SitesChecks(Tx, vector[vector[string]], bool, Tx) except +
        
        void initialise_choices()
        Chooser * __getitem__(string) except +
        
        void check_position(int)
        string check_consequence(string, string, int)
    
    cdef Region _get_gene_range(Tx)
    cdef string _get_mutated_aa(Tx, string, string, int) except +

cdef class SiteRates:
    cdef SitesChecks *_checks  # hold a C++ instance which we're wrapping
    def __cinit__(self, Transcript transcript, vector[vector[string]] rates,
            Transcript masked_sites=None, cds_coords=True):
        
        if transcript is None:
            raise ValueError('no transcript supplied')
        
        if masked_sites is None:
            self._checks = new SitesChecks(deref(transcript.thisptr), rates,
                cds_coords)
        else:
            self._checks = new SitesChecks(deref(transcript.thisptr), rates,
                cds_coords, deref(masked_sites.thisptr))
    
    def __dealloc__(self):
        del self._checks
    
    def __getitem__(self, category):
        ''' get site-specific mutation rates for each CDS base
    
        Args:
            category: string to indicate the consequence type. The permitted
                types are "missense", "nonsense", "synonymous",
                "loss_of_function", "splice_lof", and "splice_region".
        
        Returns:
            A WeightedChoice object for the CDS, where each position is paired
            with its mutation rate. We can then randomly sample sites from the
            CDS WeightedChoice object according to the probability of each site
            being mutated to the specific consequence type.
        '''
        
        cdef Chooser * chooser = self._checks.__getitem__(category.encode('utf8'))
        
        choices = WeightedChoice()
        choices.thisptr.append(deref(chooser))
        
        return choices
    
    def clear(self):
        self._checks.initialise_choices()
    
    def check_position(self, bp):
        self._checks.check_position(bp)
    
    def check_consequence(self, initial_aa, mutated_aa, position):
        initial_aa = initial_aa.encode('utf8')
        mutated_aa = mutated_aa.encode('utf8')
        return self._checks.check_consequence(initial_aa, mutated_aa, position).decode('utf8')

def get_gene_range(Transcript tx):
    region = _get_gene_range(deref(tx.thisptr))
    return {"start": region.start, "end": region.end}

def get_mutated_aa(Transcript tx,  base, codon, intra_codon):
    
    base = base.encode('utf8')
    codon = codon.encode('utf8')
    
    return _get_mutated_aa(deref(tx.thisptr), base, codon, intra_codon).decode('utf8')



""" class to test the EnsemblRequest class
"""

import json
import os
from pathlib import Path
import asyncio
import unittest
import time
import tempfile
import shutil

from denovonear.rate_limiter import RateLimiter
from denovonear.ensembl_requester import (get_genes_for_hgnc_id,
    get_transcript_ids_for_ensembl_gene_id, get_previous_symbol,
    get_genomic_seq_for_transcript, get_cds_seq_for_transcript,
    get_protein_seq_for_transcript, get_exon_ranges_for_transcript,
    get_cds_ranges_for_transcript)

async def call(func, *args, **kwargs):
    ''' call ensembl rest API function
    '''
    async with RateLimiter(15) as ensembl:
        return await func(ensembl, *args, **kwargs)

def _run(func, *args, **kwargs):
    return asyncio.get_event_loop().run_until_complete(call(func, *args, **kwargs))

class TestEnsemblRequestPy(unittest.TestCase):
    """ unit test the EnsemblRequest class
    """
    
    def test_get_genes_for_hgnc_id(self):
        """ test that get_genes_for_hgnc_id() works correctly
        """
        
        genes = _run(get_genes_for_hgnc_id, "KMT2A", build='grch37')
        self.assertEqual(genes, ['ENSG00000118058', 'ENSG00000267910'])
    
    def test_get_previous_symbol(self):
        """ test that get_previous_symbol() works correctly
        """
        
        prev = _run(get_previous_symbol, "KMT2A")
        self.assertEqual(prev, ["MLL"])
        
        # make a check for a gene with multiple documents, to check that we
        # don't raise an error
        prev = _run(get_previous_symbol, "KRT16P1")
        self.assertEqual(prev, ["KRT14P"])
    
    def test_get_transcript_ids_for_ensembl_gene_id(self):
        """ test that get_transcript_ids_for_ensembl_gene_id() works correctly
        """
        
        hgnc = ["KMT2A", "MLL"]
        
        enst = _run(get_transcript_ids_for_ensembl_gene_id, 'ENSG00000118058', hgnc)
        enst += _run(get_transcript_ids_for_ensembl_gene_id, 'ENSG00000267910', hgnc)
        
        self.assertEqual(set(enst), set(['ENST00000534358', 'ENST00000531904',
            'ENST00000389506', 'ENST00000354520', 'ENST00000532204',
            'ENST00000529852', 'ENST00000527869', 'ENST00000533790',
            'ENST00000392873']))
    
    def test_get_genomic_seq_for_transcript(self):
        """ check that get_genomic_seq_for_transcript() works correctly
        """
        
        seq = _run(get_genomic_seq_for_transcript, "ENST00000302030", expand=0)
        
        self.assertEqual(seq, ('11', 59210617, 59211667, '+', 'CTTGTCCTTGTGGTCC'
            'ACGGGAAGCATGTCCATAACCAAAGCCTGGAACAGCTCATCAGTGACCATGTTCATCCTCCTGGGA'
            'TTCACAGACCATCCAGAACTCCAGGCCCTCCTCTTTGTGACCTTCCTGGGCATCTATCTTACCACC'
            'CTGGCCTGGAACCTGGCCCTCATTTTTCTGATCAGAGGTGACACCCATCTGCACACACCCATGTAC'
            'TTCTTCCTAAGCAACTTATCTTTCATTGACATCTGCTACTCTTCTGCTGTGGCTCCCAATATGCTC'
            'ACTGACTTCTTCTGGGAGCAGAAGACCATATCATTTGTGGGCTGTGCTGCTCAGTTTTTTTTCTTT'
            'GTCGGCATGGGTCTGTCTGAGTGCCTCCTCCTGACTGCTATGGCATACGACCGATATGCAGCCATC'
            'TCCAGCCCCCTTCTCTACCCCACTATCATGACCCAGGGCCTCTGTACACGCATGGTGGTTGGGGCA'
            'TATGTTGGTGGCTTCCTGAGCTCCCTGATCCAGGCCAGCTCCATATTTAGGCTTCACTTTTGCGGA'
            'CCCAACATCATCAACCACTTCTTCTGCGACCTCCCACCAGTCCTGGCTCTGTCTTGCTCTGACACC'
            'TTCCTCAGTCAAGTGGTGAATTTCCTCGTGGTGGTCACTGTCGGAGGAACATCGTTCCTCCAACTC'
            'CTTATCTCCTATGGTTACATAGTGTCTGCGGTCCTGAAGATCCCTTCAGCAGAGGGCCGATGGAAA'
            'GCCTGCAACACGTGTGCCTCGCATCTGATGGTGGTGACTCTGCTGTTTGGGACAGCCCTTTTCGTG'
            'TACTTGCGACCCAGCTCCAGCTACTTGCTAGGCAGGGACAAGGTGGTGTCTGTTTTCTATTCATTG'
            'GTGATCCCCATGCTGAACCCTCTCATTTACAGTTTGAGGAACAAAGAGATCAAGGATGCCCTGTGG'
            'AAGGTGTTGGAAAGGAAGAAAGTGTTTTCTTAGGTCATGCGTAGAAACTTATTTATCCAAACTGCT'
            'GGAGAATTAAACAATCCAAGCCTTCACCTCCACCTCTGCCTCAGG'))
    
    def test_get_cds_seq_for_transcript(self):
        """ check that get_cds_seq_for_transcript() works correctly
        """
        
        seq = _run(get_cds_seq_for_transcript, "ENST00000302030")
        self.assertEqual(seq, 'ATGTCCATAACCAAAGCCTGGAACAGCTCATCAGTGACCATGTTCATC'
            'CTCCTGGGATTCACAGACCATCCAGAACTCCAGGCCCTCCTCTTTGTGACCTTCCTGGGCATCTAT'
            'CTTACCACCCTGGCCTGGAACCTGGCCCTCATTTTTCTGATCAGAGGTGACACCCATCTGCACACA'
            'CCCATGTACTTCTTCCTAAGCAACTTATCTTTCATTGACATCTGCTACTCTTCTGCTGTGGCTCCC'
            'AATATGCTCACTGACTTCTTCTGGGAGCAGAAGACCATATCATTTGTGGGCTGTGCTGCTCAGTTT'
            'TTTTTCTTTGTCGGCATGGGTCTGTCTGAGTGCCTCCTCCTGACTGCTATGGCATACGACCGATAT'
            'GCAGCCATCTCCAGCCCCCTTCTCTACCCCACTATCATGACCCAGGGCCTCTGTACACGCATGGTG'
            'GTTGGGGCATATGTTGGTGGCTTCCTGAGCTCCCTGATCCAGGCCAGCTCCATATTTAGGCTTCAC'
            'TTTTGCGGACCCAACATCATCAACCACTTCTTCTGCGACCTCCCACCAGTCCTGGCTCTGTCTTGC'
            'TCTGACACCTTCCTCAGTCAAGTGGTGAATTTCCTCGTGGTGGTCACTGTCGGAGGAACATCGTTC'
            'CTCCAACTCCTTATCTCCTATGGTTACATAGTGTCTGCGGTCCTGAAGATCCCTTCAGCAGAGGGC'
            'CGATGGAAAGCCTGCAACACGTGTGCCTCGCATCTGATGGTGGTGACTCTGCTGTTTGGGACAGCC'
            'CTTTTCGTGTACTTGCGACCCAGCTCCAGCTACTTGCTAGGCAGGGACAAGGTGGTGTCTGTTTTC'
            'TATTCATTGGTGATCCCCATGCTGAACCCTCTCATTTACAGTTTGAGGAACAAAGAGATCAAGGAT'
            'GCCCTGTGGAAGGTGTTGGAAAGGAAGAAAGTGTTTTCTTAG')
    
    def test_get_protein_seq_for_transcript(self):
        """ test that get_protein_seq_for_transcript() works correctly
        """
        
        seq = _run(get_protein_seq_for_transcript, "ENST00000302030")
        self.assertEqual(seq, 'MSITKAWNSSSVTMFILLGFTDHPELQALLFVTFLGIYLTTLAWNLAL'
            'IFLIRGDTHLHTPMYFFLSNLSFIDICYSSAVAPNMLTDFFWEQKTISFVGCAAQFFFFVGMGLSE'
            'CLLLTAMAYDRYAAISSPLLYPTIMTQGLCTRMVVGAYVGGFLSSLIQASSIFRLHFCGPNIINHF'
            'FCDLPPVLALSCSDTFLSQVVNFLVVVTVGGTSFLQLLISYGYIVSAVLKIPSAEGRWKACNTCAS'
            'HLMVVTLLFGTALFVYLRPSSSYLLGRDKVVSVFYSLVIPMLNPLIYSLRNKEIKDALWKVLERKK'
            'VFS')
    
    def test_get_exon_ranges_for_transcript(self):
        """ test that get_exon_ranges_for_transcript() works correctly
        """
        
        exons = _run(get_exon_ranges_for_transcript, "ENST00000534358")
        self.assertEqual(exons, [(118307205, 118307659), (118339490, 118339559),
            (118342377, 118345030), (118347520, 118347697), (118348682, 118348916),
            (118350889, 118350953), (118352430, 118352807), (118353137, 118353210),
            (118354898, 118355029), (118355577, 118355690), (118359329, 118359475),
            (118360507, 118360602), (118360844, 118360964), (118361911, 118362033),
            (118362459, 118362643), (118363772, 118363945), (118365003, 118365113),
            (118365409, 118365482), (118366415, 118366608), (118366976, 118367082),
            (118368651, 118368788), (118369085, 118369243), (118370018, 118370135),
            (118370550, 118370628), (118371702, 118371862), (118372387, 118372572),
            (118373113, 118377361), (118378244, 118378324), (118379851, 118379915),
            (118380663, 118380833), (118382666, 118382740), (118390333, 118390507),
            (118390672, 118390779), (118391517, 118391600), (118392003, 118392132),
            (118392612, 118397539)])
    
    def test_get_cds_ranges_for_transcript(self):
        """ tets that get_cds_ranges_for_transcript() works correctly
        """
        
        cds = _run(get_cds_ranges_for_transcript, "ENST00000534358")
        self.assertEqual(cds, [(118307228, 118307659), (118339490, 118339559),
            (118342377, 118345030), (118347520, 118347697), (118348682, 118348916),
            (118350889, 118350953), (118352430, 118352807), (118353137, 118353210),
            (118354898, 118355029), (118355577, 118355690), (118359329, 118359475),
            (118360507, 118360602), (118360844, 118360964), (118361911, 118362033),
            (118362459, 118362643), (118363772, 118363945), (118365003, 118365113),
            (118365409, 118365482), (118366415, 118366608), (118366976, 118367082),
            (118368651, 118368788), (118369085, 118369243), (118370018, 118370135),
            (118370550, 118370628), (118371702, 118371862), (118372387, 118372572),
            (118373113, 118377361), (118378244, 118378324), (118379851, 118379915),
            (118380663, 118380833), (118382666, 118382740), (118390333, 118390507),
            (118390672, 118390779), (118391517, 118391600), (118392003, 118392132),
            (118392612, 118392887)])



# cython: language_level=3, boundscheck=False
"""
Copyright (c) 2015 Genome Research Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from libcpp.vector cimport vector
from libcpp cimport bool
from cython.operator cimport dereference as deref

cdef class WeightedChoice:
    def __cinit__(self):
        self.thisptr = new Chooser()
        self.pos = 0
    
    def __dealloc__(self):
        if self.thisptr is not NULL:
            del self.thisptr
    
    def __len__(self):
        return self.thisptr.len()
    
    def __iter__(self):
        return self
    
    def __next__(self):
        if self.pos >= len(self):
            self.pos = 0
            raise StopIteration
        
        iter = self.thisptr.iter(self.pos)
        self.pos += 1
        
        return {"pos": iter.pos, "ref": iter.ref.decode('utf8'),
            "alt": iter.alt.decode('utf8'), 'prob': iter.prob, "offset": iter.offset}
    
    def append(self, WeightedChoice other):
        ''' combines the sites from two WeightedChoice objects
    
        NOTE: This is inefficient, but I can't figure out an easier way around
        NOTE: cython.
    
        Args:
            other: WeightedChoice object
        '''
    
        self.thisptr.append(deref(other.thisptr))
    
    def add_choice(self, site, prob, ref='N', alt='N', offset=0):
        """ add another possible choice for selection
        
        Args:
            site: a CDS position of the selected base.
            prob: probability of selecting this base.
            ref: string for reference allele
            alt: string for alternate allele
        """
        
        ref = ref.encode('utf8')
        alt = alt.encode('utf8')
        
        if len(ref) > 1 or len(alt) > 1:
            raise TypeError("requires single base alleles: {}, {}".format(ref, alt))
        
        self.thisptr.add_choice(site, prob, ref, alt, offset)
    
    def choice(self):
        """ chooses a random element using a set of probability weights
        
        Returns:
            the name of the randomly selected element (e.g. position)
        """
        
        return self.choice_with_alleles()["pos"]
    
    def choice_with_alleles(self):
        """ chooses a random element, but include alleles in output
        
        Returns:
            a dictionary of the randomly selected element, with "pos", "ref"
            and "alt" entries.
        """
        
        choice = self.thisptr.choice()
        
        return {"pos": choice.pos, "ref": choice.ref.decode('utf8'),
            "alt": choice.alt.decode('utf8'), "offset": choice.offset}
    
    def get_summed_rate(self):
        """ return the cumulative probability for the object
        """
        return self.thisptr.get_summed_rate()

cdef extern from "simulate.h":
    vector[int] _get_distances(vector[int])
    bool _has_zero(vector[int])
    double _geomean(vector[int])
    bool _halt_permutation(double, int, double, double)
    vector[double] _simulate_distribution(Chooser, int, int)
    double _analyse_de_novos(Chooser, int, int, double)

def get_distances(vector[int] positions):
    """ gets the distances between two or more CDS positions
    
    Args:
        positions: list of CDS positions as ints
    
    Returns:
        list of pairwise distances between sites
    """
    
    return _get_distances(positions)

def has_zero(vector[int] distances):
    """ figure out whether any of the pairwise distances is zero
    """
    
    return _has_zero(distances)

def geomean(vector[int] distances):
    """ gets the geometric mean distance between two or more CDS positions
    
    Args:
        distances: list of distances between CDS positions
    
    Returns:
        provides the mean distance of the pairwise distances
    """
    
    return _geomean(distances)

def simulate_distribution(WeightedChoice choices, int iterations, int de_novos_count):
    """
    """
    
    return _simulate_distribution(deref(choices.thisptr), iterations, de_novos_count)

def analyse_de_novos(WeightedChoice choices, int iterations, int de_novos_count, double observed_value):
    """
    """
    
    return _analyse_de_novos(deref(choices.thisptr), iterations, de_novos_count, observed_value)



""" load trincleotide mutation rates
"""

from pkg_resources import resource_filename

def load_mutation_rates(path=None):
    """ load sequence context-based mutation rates
    
    Args:
        path: path to table of sequence context-based mutation rates. If None,
            this defaults to per-trinucleotide rates provided by Kaitlin Samocha
            (Broad Institute).
    
    Returns:
        list of [initial, changed, rate] lists e.g. [['AGA', 'ATA', '5e-8']]
    """
    
    if path is None:
        path = resource_filename(__name__, "data/rates.txt")
    
    rates = []
    with open(path) as handle:
        for line in handle:
            if line.startswith("from"): # ignore the header line
                continue
            
            line = [ x.encode('utf8') for x in line.strip().split() ]
            rates.append(line)
    
    return rates



""" Script to investigate the probability of multiple mutations clustering
within a single gene.
"""

import os
import sys
import asyncio
import argparse
import logging

from denovonear.rate_limiter import RateLimiter
from denovonear.load_mutation_rates import load_mutation_rates
from denovonear.load_de_novos import load_de_novos
from denovonear.cluster_test import cluster_de_novos

from denovonear.load_gene import (construct_gene_object,
    count_de_novos_per_transcript, minimise_transcripts)
from denovonear.site_specific_rates import SiteRates
from denovonear.frameshift_rate import include_frameshift_rates
from denovonear.log_transform_rates import log_transform

async def clustering(ensembl, mut_dict, output, args):
    
    de_novos = load_de_novos(args.input)
    
    output.write("gene_id\tmutation_category\tevents_n\tdist\tprobability\n")
    
    iterations = 1000000
    for symbol in sorted(de_novos):
        
        if len(de_novos[symbol]["missense"] + de_novos[symbol]["nonsense"]) < 2:
            continue
        
        probs = await cluster_de_novos(symbol, de_novos[symbol], ensembl, iterations, mut_dict)
        
        if probs is None:
            continue
        
        output.write("{}\t{}\t{}\t{}\t{}\n".format(symbol, "missense",
            len(de_novos[symbol]["missense"]), probs["miss_dist"], probs["miss_prob"]))
        output.write("{}\t{}\t{}\t{}\t{}\n".format(symbol, "nonsense",
            len(de_novos[symbol]["nonsense"]), probs["nons_dist"], probs["nons_prob"]))

async def find_transcripts(ensembl, mut_dict, output, args):
    
    de_novos = load_de_novos(args.de_novos)
    
    output.write("hgnc_symbol\ttranscript_id\tlength\tde_novos\n")
    
    for symbol in sorted(de_novos):
        print(symbol)
        func_events = de_novos[symbol]["missense"] + de_novos[symbol]["nonsense"]
        
        # find the counts per transcript, depending on whether we want to count
        # for all transcripts containing one or more de novos, or to find the
        # minimum set of transcripts to contain the de novos
        try:
            if args.all_transcripts:
                counts = await count_de_novos_per_transcript(ensembl, symbol, func_events)
            elif args.minimal_transcripts:
                counts = await minimise_transcripts(ensembl, symbol, func_events)
        except (ValueError, IndexError):
            print("error occured with {0}".format(symbol))
            continue
        
        # write the transcript details to a file
        for key in counts:
            line = "{}\t{}\t{}\t{}\n".format(symbol, key, counts[key]["len"],
                counts[key]["n"])
            output.write(line)

def load_genes(path):
    """ load a file listing gene and transcript IDs
    
    Args:
        path: path to file containing gene IDs and transcript IDs e.g.
            gene_1    transcript_1.1    length_1    denovo_count
            gene_2    transcript_2.1    length_3    denovo_count
    
    Returns:
        dict of transcripts eg {'CTC1': ["ENST00000315684", "ENST00000485511"]}
    """
    
    with open(path, 'rt') as f:
        lines = [ x.split('\t')[:2] for x in f if not x.startswith('hgnc') ]
    
    transcripts = {}
    for symbol, tx in lines:
        if symbol not in transcripts:
            transcripts[symbol] = []
        transcripts[symbol].append(tx)
    
    return transcripts

async def get_mutation_rates(transcripts, mut_dict, ensembl):
    """ determines mutation rates per functional category for transcripts
    
    Args:
        transcripts: list of transcript IDs for a gene
        mut_dict: dictionary of local sequence context mutation rates
        ensembl: EnsemblRequest object, to retrieve information from Ensembl.
    
    Returns:
        tuple of (rates, merged transcript, and transcript CDS length)
    """
    
    rates = {'missense': 0, 'nonsense': 0, 'splice_lof': 0,
        'splice_region': 0, 'synonymous': 0}
    combined = None
    
    for tx_id in transcripts:
        try:
            tx = await construct_gene_object(ensembl, tx_id)
        except ValueError:
            continue
        
        if len(tx.get_cds_sequence()) % 3 != 0:
            raise ValueError("anomalous_coding_sequence")
        
        # ignore mitochondrial genes
        if tx.get_chrom() == "MT":
            continue
        
        sites = SiteRates(tx, mut_dict, masked_sites=combined)
        combined = tx + combined
        
        for cq in ['missense', 'nonsense', 'splice_lof', 'splice_region', 'synonymous']:
            rates[cq] += sites[cq].get_summed_rate()
    
    if combined is None:
        raise ValueError('no tx found')
    
    length = combined.get_coding_distance(combined.get_cds_end())['pos']
    
    return rates, combined, length

async def gene_rates(ensembl, mut_dict, output, args):
    
    transcripts = load_genes(args.genes)
    
    header = ['transcript_id', 'chrom', 'length', 'missense_rate', 'nonsense_rate',
        'splice_lof_rate', 'splice_region_rate', 'synonymous_rate']
    output.write('\t'.join(header) + "\n")
    
    for symbol in sorted(transcripts):
        print(symbol)
        try:
            rates, tx, length = await get_mutation_rates(transcripts[symbol],
                mut_dict, ensembl)
            # log transform rates, for consistency with Samocha et al.
            line = "{}\t{}\t{}\t{}".format(symbol, tx.get_chrom(), length, log_transform(rates))
        except (ValueError, KeyError) as error:
            print("{}\t{}\n".format(symbol, error))
            line = "{}\t{}\tNA\tNA\tNA\tNA\tNA\tNA".format(symbol, tx.get_chrom())
        
        output.write(line + '\n')
    
    output.close()
    include_frameshift_rates(args.out)

def get_options():
    """ get the command line switches
    """
    
    parser = argparse.ArgumentParser(description='denovonear cli interface')
    
    ############################################################################
    # CLI options in common
    parent = argparse.ArgumentParser(add_help=False)
    parent.add_argument("--out", help="output filename")
    parent.add_argument("--rates",
        help="Path to file containing sequence context-based mutation rates.")
    parent.add_argument("--genome-build", choices=["grch37",
        "GRCh37", "grch38", "GRCh38"], default="grch37", help="Genome build "
        "that the de novo coordinates are based on (GRCh37 or GRCh38")
    parent.add_argument("--log", default='ensembl_requests.log', help="where to write log files")
    
    subparsers = parser.add_subparsers()
    
    ############################################################################
    # CLI options for clustering
    cluster = subparsers.add_parser('cluster', parents=[parent],
        description="Tests the proximity of de novo mutations in genes.")
    cluster.add_argument("--in", dest="input", required=True, help="Path to "
        "file listing known mutations in genes. See example file in data folder "
        "for format.")
    
    cluster.set_defaults(func=clustering)
    
    ############################################################################
    # CLI options for identifing transcripts to use
    transcripts = subparsers.add_parser('transcripts', parents=[parent],
        description="Identify transcripts for a gene containing de novo events.")
    transcripts.add_argument("--de-novos", required=True, help="Path to "
        "file listing de novo variants in genes.")
    transcripts.set_defaults(func=find_transcripts)
    group = transcripts.add_mutually_exclusive_group(required=True)
    group.add_argument("--all-transcripts", action="store_true", default=False,
        help="Flag if you want to identify all transcripts with more than "
        "one de novo on it.")
    group.add_argument("--minimise-transcripts", action="store_true",
        default=False, help="Flag if you want to identify the minimal set of "
        "transcripts to contain all de novos.")
    
    ############################################################################
    # CLI options for identifing transcripts to use
    rater = subparsers.add_parser("rates", parents=[parent],
        description="determine mutation rates for genes given transcript IDs.")
    rater.add_argument("--genes", help="Path to file "
        "listing HGNC symbols, with one or more transcript IDs per gene. "
        "The tab-separated input format is gene symbol followed by transcript "
        "ID. Alternative transcripts are listed on separate lines.")
    rater.set_defaults(func=gene_rates)
    
    args = parser.parse_args()
    if 'func' not in args:
        print('Use one of the subcommands: cluster, rates, or transcripts\n')
        parser.print_help()
        sys.exit()
    
    return args

async def runner():
    args = get_options()
    FORMAT = '%(asctime)-15s %(message)s'
    logging.basicConfig(filename=args.log, format=FORMAT, level=logging.INFO)
    
    async with RateLimiter(per_second=15) as ensembl:
        mut_dict = load_mutation_rates(args.rates)
        with open(args.out, "wt") as output:
            await args.func(ensembl, mut_dict, output, args)

def main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(runner())

if __name__ == '__main__':
    main()



"""
Copyright (c) 2015 Genome Research Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import unittest
import tempfile
import os
import sys
import shutil
import zlib
from datetime import datetime, timedelta
from threading import Thread, Event
import random
import hashlib
import time

from denovonear.ensembl_cache import EnsemblCache

IS_PYTHON2 = sys.version_info[0] == 2
IS_PYTHON3 = sys.version_info[0] == 3

class TestEnsemblCachePy(unittest.TestCase):
    """ unit test the EnsemblCache class
    """
    
    @classmethod
    def setUpClass(self):
        self.temp_dir = tempfile.mkdtemp()
        self.cache = EnsemblCache(self.temp_dir)
    
    @classmethod
    def tearDownClass(self):
        shutil.rmtree(self.temp_dir)
    
    def test_get_key_from_url(self):
        """ test that get_key_from_url() works correctly
        """
        
        self.assertEqual(self.cache.get_key_from_url("http://rest.ensembl.org/info/rest"), ("info.rest", 'grch38'))
        self.assertEqual(self.cache.get_key_from_url("http://rest.ensembl.org/xrefs/symbol/homo_sapiens/ABO"), ("xrefs.symbol.homo_sapiens.ABO", 'grch38'))
        self.assertEqual(self.cache.get_key_from_url("http://rest.ensembl.org/sequence/id/ENST00000378520?type=protein"), ("sequence.id.ENST00000378520.protein", 'grch38'))
        self.assertEqual(self.cache.get_key_from_url("http://rest.ensembl.org/feature/id/ENSG00000175164?feature=transcript"), ("feature.id.ENSG00000175164.transcript", 'grch38'))
        self.assertEqual(self.cache.get_key_from_url("http://rest.ensembl.org/sequence/id/ENST00000538324?type=genomic;expand_3prime=10;expand_5prime=10"), ("sequence.id.ENST00000538324.genomic", 'grch38'))
        self.assertEqual(self.cache.get_key_from_url("http://rest.ensembl.org/sequence/id/ENST00000538324?type=cds"), ("sequence.id.ENST00000538324.cds", 'grch38'))
        self.assertEqual(self.cache.get_key_from_url("http://rest.ensembl.org/feature/id/ENST00000538324?feature=exon"), ("feature.id.ENST00000538324.exon", 'grch38'))
        self.assertEqual(self.cache.get_key_from_url("http://rest.ensembl.org/vep/human/id/rs3887873/consequences?"), ("vep.human.id.rs3887873.consequences", 'grch38'))
        self.assertEqual(self.cache.get_key_from_url("http://rest.ensembl.org/vep/human/9:22125503-22125502:1/C/consequences?"), ("vep.human.9_22125503-22125502_1.C.consequences", 'grch38'))
        self.assertEqual(self.cache.get_key_from_url("http://grch37.rest.ensembl.org/vep/human/9:22125503-22125502:1/C/consequences?"), ("vep.human.9_22125503-22125502_1.C.consequences", 'grch37'))
    
    def test_get_cached_data(self):
        """ test that get_cached_data() works correctly
        """
        
        # set up the data to go in the database
        url = "http://rest.ensembl.org/feature/id/temp1?feature=exon"
        string = b"temp_data"
        
        # check that the data is not in the database to start
        self.assertIsNone(self.cache.get_cached_data(url))
        
        # insert the data in the database
        self.cache.cache_url_data(url, string)
        
        # check that some data is now in the database
        data = self.cache.get_cached_data(url)
        self.assertIsNotNone(data)
        
        # check that the data is correct if the row is in the database
        self.assertEqual(data, string)
    
    def test_get_cached_data_old_date(self):
        """ check that the cache ignores outdated data
        """
        url = "http://rest.ensembl.org/feature/id/temp1?feature=exon"
        string = b"temp_data"
        today = datetime.today()
        long_ago = today - timedelta(days=181)
        
        # check that obsolete data returns False
        self.cache.today = long_ago
        self.cache.cache_url_data(url, string)
        self.assertIsNotNone(self.cache.get_cached_data(url))
        
        self.cache.today = today
        self.assertIsNone(self.cache.get_cached_data(url))
    
    def test_cache_url_data(self):
        """ test that cache_url_data works correctly
        """
        
        # set up the data to go in the database
        url = "http://rest.ensembl.org/feature/id/temp2?feature=exon"
        temp_data = b"temp_data"
        
        # check that the data is not in before we insert it
        self.assertIsNone(self.cache.get_cached_data(url))
        
        # insert the data, then check that it has gone in
        self.cache.cache_url_data(url, temp_data)
        self.assertIsNotNone(self.cache.get_cached_data(url))
    
    def test_cache_load(self):
        """ make sure the cache can handle a reasonable load
    
        This test uses multiple threads writing to the cache simultaneously to
        show the cache can handle the load. Failure is shown by an exception.
        """
    
        cache_dir = os.path.join(self.temp_dir, 'loading')
        os.mkdir(cache_dir)
        text = lambda l: '{:x}'.format(random.getrandbits(l * 4)).strip().encode('utf8')
        url = lambda : 'example.com/base/sub/{}'.format(text(10))
        write = lambda cache: cache.cache_url_data(url(), text(100))
        
        class Runner(Thread):
            def __init__(self, counter=100):
                super(Runner, self).__init__()
                self.counter = counter
            def run(self):
                cache = EnsemblCache(cache_dir)
                while self.counter > 0:
                    write(cache)
                    self.counter -= 1
        
        try:
            threads = [ Runner() for x in range(50) ]
            [ x.start() for x in threads ]
            [ x.join() for x in threads ]
        except:
            self.fail("EnsemblCache failed under heavy load")



# cython: language_level=3, boundscheck=False
"""
Copyright (c) 2015 Genome Research Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from libcpp.string cimport string

cdef extern from "weighted_choice.h":
    cdef cppclass Chooser:
        Chooser() except +
        void add_choice(int, double, string, string, int)
        AlleleChoice choice()
        double get_summed_rate()
        int len()
        AlleleChoice iter(int)
        void append(Chooser)
    
    cdef struct AlleleChoice:
        int pos
        string ref
        string alt
        double prob
        int offset

cdef class WeightedChoice:
    cdef int pos
    cdef Chooser *thisptr # hold a C++ instance which we're wrapping



"""
Copyright (c) 2015 Genome Research Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import math
import unittest

from denovonear.weights import get_distances, geomean

class TestGeomeanPy(unittest.TestCase):
    """ unit test the geomean function
    """
    
    def test_geomean(self):
        """ test geomean() works correctly
        """
        
        self.assertEqual(geomean([0]), 0)
        self.assertEqual(geomean([1]), 1)
        self.assertEqual(geomean([1, 1]), 1)
        self.assertEqual(geomean([1, 2]), 1.4142135623730951)
        self.assertEqual(geomean([10]), 10)
        self.assertEqual(geomean([0, 900, 900]), 92.2860120092046)
        self.assertEqual(geomean([10, 890, 900]), 200.08329863520368)
        
        # check that if we try to get the mean distance for a list with fewer
        # than two positions in it, we get an error, since we can't estimate any
        # distances
        self.assertTrue(math.isnan(geomean([])))
    
    def test_get_distance(self):
        """ test that get_distances() works correctly
        """
        
        # check pairwise distances
        self.assertEqual(get_distances([0, 0]), [0])
        self.assertEqual(get_distances([0, 1]), [1])
        self.assertEqual(get_distances([0, 1, 10]), [1, 10, 9])
        
        # empty lists (or lists with one entry) can't give pairwise distances
        self.assertEqual(get_distances([]), [])
        self.assertEqual(get_distances([0]), [])
        
        # check that float positionss are automatically converted to integers
        self.assertEqual(get_distances([0.5, 10]), [10])
        
        # check that negative positions still work
        self.assertEqual(get_distances([-10, -5]), [5])
        
        # and check we get an error if we call with strings
        with self.assertRaises(TypeError):
            get_distances([0, 1, "e"])
        
    



""" class to test the Transcript class
"""

import unittest

from denovonear.transcript import Transcript

class TestTranscriptPy(unittest.TestCase):
    """ unit test the Transcript class
    """
    
    def setUp(self):
        """ construct a Transcript object for unit tests
        """
        
        self.gene = self.construct_gene()
    
    def construct_gene(self, name='TEST', chrom='1', start=1000, end=2000,
            strand='+', exons=[(1000, 1200), (1800, 2000)],
            cds=[(1100, 1200), (1800, 1900)]):
        
        tx = Transcript(name, chrom, start, end, strand)
        tx.set_exons(exons, cds)
        tx.set_cds(cds)
        
        return tx
    
    def test_set_exons(self):
        """ test that set_exons() works correctly
        """
        
        exons = [(0, 200), (800, 1000)]
        cds = [(1100, 1200), (1800, 1900)]
        self.gene.set_exons(exons, cds)
        
        self.assertEqual(self.gene.get_exons(), [{'start': 0, 'end': 200}, {'start': 800, 'end': 1000}])
        
        self.gene = self.construct_gene(strand='-')
        self.gene.set_exons(exons, cds)
        self.assertEqual(self.gene.get_exons(), [{'start': 0, 'end': 200}, {'start': 800, 'end': 1000}])
    
    def test_set_exons_missing_exon(self):
        """ test that set_exons() works correctly when we lack coordinates
        """
        
        # also test that we can determine the exons when we don't have any given
        # for a transcript with a single CDS region
        exons = []
        cds = [(1100, 1200)]
        self.gene.set_exons(exons, cds)
        self.assertEqual(self.gene.get_exons(), [{'start': 1000, 'end': 2000}])
        
        # check that missing exons, but 2+ CDS regions raises an error.
        cds = [(1100, 1200), (1300, 1400)]
        with self.assertRaises(ValueError):
            self.gene.set_exons(exons, cds)
    
    def test_set_cds(self):
        """ test that set_cds() works correctly
        """
        
        exons = [(0, 200), (800, 1000)]
        cds = [(100, 200), (800, 900)]
        
        # make sure we raise an error if we try to set the CDS before the exons
        with self.assertRaises(ValueError):
            tx = Transcript('test', '1', 0, 1000, '+')
            tx.set_cds(cds)
        
        # check CDS positions
        self.gene.set_exons(exons, cds)
        self.gene.set_cds(cds)
        self.assertEqual(self.gene.get_cds(), [{'start': 100, 'end': 200}, {'start': 800, 'end': 900}])
        
        # check that CDS ends outside an exon are corrected
        exons = [(0, 200), (300, 400), (800, 1000)]
        cds = [(100, 200), (300, 402)]
        self.gene.set_exons(exons, cds)
        self.gene.set_cds(cds)
        self.assertEqual(self.gene.get_cds(), [{'start': 100, 'end': 200},
            {'start': 300, 'end': 400}, {'start': 800, 'end': 802}])
        
        cds = [(298, 400), (800, 1000)]
        self.gene.set_exons(exons, cds)
        self.gene.set_cds(cds)
        self.assertEqual(self.gene.get_cds(), [{'start': 198, 'end': 200},
            {'start': 300, 'end': 400}, {'start': 800, 'end': 1000}])
    
    def test_fix_cds_boundary(self):
        """ test that _fix_out_of_exon_cds_boundary() works correctly
        """
        
        exons = [(1100, 1200), (1300, 1400), (1800, 1900)]
        cds = [(1300, 1400)]
        
        tx = Transcript('test', '1', 0, 1000, '+')
        
        tx.set_exons(exons, cds)
        
        self.assertEqual(tx.fix_cds_boundary(1295), {'start': 1195, 'end': 1200})
        self.assertEqual(tx.fix_cds_boundary(1205), {'start': 1300, 'end': 1305})
        
        self.assertEqual(tx.fix_cds_boundary(1402), {'start': 1800, 'end': 1802})
        self.assertEqual(tx.fix_cds_boundary(1798), {'start': 1398, 'end': 1400})
        
        # raise an error if the position is within the exons
        with self.assertRaises(ValueError):
            self.gene.fix_cds_boundary(1105)
    
    def test_get_cds_range(self):
        """ unit test checking the CDS end points by strand
        """
        
        # self.gene.cds_min = 100
        # self.gene.cds_max = 200
        
        self.gene = self.construct_gene(exons=[(0, 300)], cds=[(100, 200)])
        
        self.assertEqual(self.gene.get_cds_start(), 100)
        self.assertEqual(self.gene.get_cds_end(), 200)
        
        self.gene = self.construct_gene(exons=[(0, 300)], cds=[(100, 200)], strand='-')
        self.assertEqual(self.gene.get_cds_start(), 200)
        self.assertEqual(self.gene.get_cds_end(), 100)
        
        with self.assertRaises(ValueError):
            self.construct_gene(strand='x')
    
    def test___add__(self):
        """ test that __add__() works correctly
        """
        
        exons = [(10, 20), (50, 60), (90, 100)]
        cds_2 = [(50, 60), (90, 95)]
        
        a = Transcript("a", "1", 10, 100, "+")
        b = Transcript("b", "1", 10, 100, "+")
        c = Transcript("c", "1", 10, 100, "+")
        d = Transcript("d", "1", 10, 100, "+")
        
        a.set_exons(exons, [(55, 60), (90, 100)])
        a.set_cds([(55, 60), (90, 100)])
        
        b.set_exons(exons, [(50, 60), (90, 95)])
        b.set_cds([(50, 60), (90, 95)])
        
        c.set_exons([(45, 65)], [(45, 65)])
        c.set_cds([(45, 65)])
        
        d.set_exons([(30, 40)], [(30, 40)])
        d.set_cds([(30, 40)])
        
        # check that adding two Transcripts gives the union of CDS regions
        self.assertEqual((a + b).get_cds(), [{'start': 50, 'end': 60}, {'start': 90, 'end': 100}])
        self.assertEqual((a + c).get_cds(), [{'start': 45, 'end': 65}, {'start': 90, 'end': 100}])
        
        # check that addition is reversible
        self.assertEqual((c + a).get_cds(), [{'start': 45, 'end': 65}, {'start': 90, 'end': 100}])
        
        # check that adding previously unknown exons works
        self.assertEqual((a + d).get_cds(), [{'start': 30, 'end': 40}, {'start': 55, 'end': 60}, {'start': 90, 'end': 100}])
        
        # check that we can add transcript + None correctly
        self.assertEqual(a + None, a)
        self.assertEqual(None + a, a)
    
    def test___add__not_overlapping(self):
        ''' test that __add__() works correctly when transcripts do not overlap
        '''
        
        a = Transcript("a", "1", 10, 50, "+")
        b = Transcript("b", "1", 60, 80, "+")
        
        a.set_exons([(10, 50)], [(10, 50)])
        a.set_cds([(10, 50)])
        a.add_genomic_sequence('N' * 40)
        
        b.set_exons([(60, 80)], [(60, 80)])
        b.set_cds([(60, 80)])
        b.add_genomic_sequence('N' * 20)
        
        self.assertEqual(len((a + b).get_genomic_sequence()), 70)
    
    def test___add__cds_length_fixed(self):
        """ check that we can merge transcripts, even with fixed CDS coords
        """
        
        a = Transcript("a", "1", 10, 20, "+")
        a.set_exons([(10, 20)], [(10, 20)])
        a.set_cds([(10, 20)])
        
        a.add_cds_sequence('ACTGTACGCAT')
        a.add_genomic_sequence('CGTAGACTGTACGCATCGATT', offset=5)
        
        b = Transcript("b", "1", 0, 10, "+")
        b.set_exons([(0, 10)], [(0, 10)])
        b.set_cds([(0, 10)])
        
        b.add_cds_sequence('ACTGTACGCAT')
        b.add_genomic_sequence('CGTAGACTGTACGCATCGTAG', offset=5)
        
        # without a fix to tx.cpp to adjust an exon coordinate simultaneously,
        # the line below would give an error.
        c = a + b
    
    def test_merge_coordinates(self):
        """ test that we can merge transcripts with odd overlaps
        """
        
        a = Transcript("a", "1", 10, 20, "+")
        
        exons1 = [{'start': 10, 'end': 20}, {'start': 25, 'end': 40}]
        exons2 = [{'start': 10, 'end': 30}]
        
        self.assertEqual(a.merge_coordinates(exons1, exons2),
            a.merge_coordinates(exons2, exons1))
    
    def test_in_exons(self):
        """ test that in_exons() works correctly
        """
        
        # self.gene.exons = [(1000, 1200), (1800, 2000)]
        
        # check for positions inside the exon ranges
        self.assertTrue(self.gene.in_exons(1000))
        self.assertTrue(self.gene.in_exons(1001))
        self.assertTrue(self.gene.in_exons(1200))
        self.assertTrue(self.gene.in_exons(1800))
        self.assertTrue(self.gene.in_exons(1801))
        self.assertTrue(self.gene.in_exons(1999))
        self.assertTrue(self.gene.in_exons(2000))
        
        # check positions outside the exon ranges
        self.assertFalse(self.gene.in_exons(999))
        self.assertFalse(self.gene.in_exons(1201))
        self.assertFalse(self.gene.in_exons(1799))
        self.assertFalse(self.gene.in_exons(2001))
        self.assertFalse(self.gene.in_exons(-1100))
    
    def test_get_closest_exon(self):
        """ test that get_closest_exon() works correctly
        """
        #
        exon_1 = {'start': 1000, 'end': 1200}
        exon_2 = {'start': 1800, 'end': 2000}
        
        # find for positions closer to the first exon
        self.assertEqual(self.gene.get_closest_exon(0), exon_1)
        self.assertEqual(self.gene.get_closest_exon(999), exon_1)
        self.assertEqual(self.gene.get_closest_exon(1000), exon_1)
        self.assertEqual(self.gene.get_closest_exon(1100), exon_1)
        self.assertEqual(self.gene.get_closest_exon(1200), exon_1)
        self.assertEqual(self.gene.get_closest_exon(1201), exon_1)
        
        # a site equidistant from the exons will pick the later exon
        self.assertEqual(self.gene.get_closest_exon(1500), exon_2)
        
        # find for positions closer to the second exon
        self.assertEqual(self.gene.get_closest_exon(1501), exon_2)
        self.assertEqual(self.gene.get_closest_exon(1799), exon_2)
        self.assertEqual(self.gene.get_closest_exon(1800), exon_2)
        self.assertEqual(self.gene.get_closest_exon(1900), exon_2)
        self.assertEqual(self.gene.get_closest_exon(2000), exon_2)
        self.assertEqual(self.gene.get_closest_exon(2001), exon_2)
        self.assertEqual(self.gene.get_closest_exon(10000), exon_2)
    
    def test_in_coding_region(self):
        """ test that in_coding_region() works correctly
        """
        
        # self.gene.cds = [(1100, 1200), (1800, 1900)]
        
        # check for positions inside the exon ranges
        self.assertTrue(self.gene.in_coding_region(1100))
        self.assertTrue(self.gene.in_coding_region(1101))
        self.assertTrue(self.gene.in_coding_region(1200))
        self.assertTrue(self.gene.in_coding_region(1800))
        self.assertTrue(self.gene.in_coding_region(1801))
        self.assertTrue(self.gene.in_coding_region(1899))
        self.assertTrue(self.gene.in_coding_region(1900))
        
        # check positions outside the exon ranges
        self.assertFalse(self.gene.in_coding_region(1099))
        self.assertFalse(self.gene.in_coding_region(1201))
        self.assertFalse(self.gene.in_coding_region(1799))
        self.assertFalse(self.gene.in_coding_region(1901))
        self.assertFalse(self.gene.in_coding_region(-1100))
    
    # def test_get_exon_containing_position(self):
    #     """ test that get_exon_containing_position() works correctly
    #     """
    #
    #     exons = [(1000, 1200), (1800, 2000)]
    #
    #     self.assertEqual(self.gene.get_exon_containing_position(1000, exons), 0)
    #     self.assertEqual(self.gene.get_exon_containing_position(1200, exons), 0)
    #     self.assertEqual(self.gene.get_exon_containing_position(1800, exons), 1)
    #     self.assertEqual(self.gene.get_exon_containing_position(2000, exons), 1)
    #
    #     # raise an error if the position isn't within the exons
    #     with self.assertRaises(RuntimeError):
    #         self.gene.get_exon_containing_position(2100, exons)
    
    def test_get_coding_distance(self):
        """ test that get_coding_distance() works correctly
        """
        
        # self.gene.cds = [(1100, 1200), (1800, 1900)]
        
        # raise an error for positions outside the CDS
        self.assertEqual(self.gene.get_coding_distance(900), {'pos': -100, 'offset': -100})
        self.assertEqual(self.gene.get_coding_distance(1000), {'pos': -100, 'offset': 0})
        self.assertEqual(self.gene.get_coding_distance(1051), {'pos': -49, 'offset': 0})
        self.assertEqual(self.gene.get_coding_distance(1300), {'pos': 100, 'offset': 100})
        self.assertEqual(self.gene.get_coding_distance(1700), {'pos': 101, 'offset': -100})
        self.assertEqual(self.gene.get_coding_distance(2000), {'pos': 301, 'offset': 0})
        self.assertEqual(self.gene.get_coding_distance(2100), {'pos': 301, 'offset': 100})
        
        # zero distance between a site and itself
        self.assertEqual(self.gene.get_coding_distance(1100), {'pos': 0, 'offset': 0})
        
        # within a single exon, the distance is between the start and end
        self.assertEqual(self.gene.get_coding_distance(1200), {'pos': 100, 'offset': 0})
        
        # if we traverse exons, the distance bumps up at exon boundaries
        self.assertEqual(self.gene.get_coding_distance(1800), {'pos': 101, 'offset': 0})
        
        # check full distance across gene
        self.assertEqual(self.gene.get_coding_distance(1900), {'pos': 201, 'offset': 0})
        
        # check that the distance bumps up for each exon boundary crossed
        cds = [(1100, 1200), (1300, 1400), (1800, 1900)]
        exons = [(1100, 1200), (1300, 1400), (1800, 1900)]
        self.gene = self.construct_gene(exons=exons, cds=cds)
        self.assertEqual(self.gene.get_coding_distance(1900), {'pos': 302, 'offset': 0})
        
        # now try a gene where the site is in an upstream exon
        self.gene = self.construct_gene(exons=[(10, 20), (30, 40), (90, 100)],
            cds= [(30, 40), (90, 95)])
        self.assertEqual(self.gene.get_coding_distance(15), {'pos': -6, 'offset': 0})
    
    def test_chrom_pos_to_cds(self):
        """ test that chrom_pos_to_cds() works correctly
        """
        # self.gene.cds = [(1100, 1200), (1800, 1900)]
        
        # note that all of these chr positions are 0-based (ie pos - 1)
        self.assertEqual(self.gene.get_coding_distance(1100), {'pos': 0, 'offset': 0})
        self.assertEqual(self.gene.get_coding_distance(1101), {'pos': 1, 'offset': 0})
        self.assertEqual(self.gene.get_coding_distance(1199), {'pos': 99, 'offset': 0})
        
        # check that outside exon boundaries gets the closest exon position, if
        # the variant is close enough
        self.assertEqual(self.gene.get_coding_distance(1200), {'pos': 100, 'offset': 0})
        self.assertEqual(self.gene.get_coding_distance(1201), {'pos': 100, 'offset': 1})
        self.assertEqual(self.gene.get_coding_distance(1798), {'pos': 101, 'offset': -2})
        self.assertEqual(self.gene.get_coding_distance(1799), {'pos': 101, 'offset': -1})
        
        # # check that sites sufficiently distant from an exon raise an error, or
        # # sites upstream of a gene, just outside the CDS, but within an exon
        # with self.assertRaises(RuntimeError):
        #     self.gene.get_coding_distance(1215)
        # with self.assertRaises(RuntimeError):
        #     self.gene.get_coding_distance(1098)
        self.assertEqual(self.gene.get_coding_distance(1215), {'pos': 100, 'offset': 15})
        self.assertEqual(self.gene.get_coding_distance(1098), {'pos': -2, 'offset': 0})
        
        # check that sites in a different exon are counted correctly
        self.assertEqual(self.gene.get_coding_distance(1799), {'pos': 101, 'offset': -1})
        
        # check that sites on the reverse strand still give the correct CDS
        self.gene = self.construct_gene(strand="-")
        self.assertEqual(self.gene.get_coding_distance(1900), {'pos': 0, 'offset': 0})
        self.assertEqual(self.gene.get_coding_distance(1890), {'pos': 10, 'offset': 0})
        self.assertEqual(self.gene.get_coding_distance(1799), {'pos': 100, 'offset': 1})
        self.assertEqual(self.gene.get_coding_distance(1792), {'pos': 100, 'offset': 8})
        self.assertEqual(self.gene.get_coding_distance(1792), {'pos': 100, 'offset': 8})
        
        self.assertEqual(self.gene.get_coding_distance(1205), {'pos': 101, 'offset': -5})
        self.assertEqual(self.gene.get_coding_distance(1200), {'pos': 101, 'offset': 0})
    




import math

def log_transform(rates):
    """ log transform a numeric value, unless it is zero, or negative
    """
    
    transformed = []
    for key in ['missense', 'nonsense', 'splice_lof', 'splice_region',
            'synonymous']:
        
        try:
            value = math.log10(rates[key])
        except ValueError:
            value = "NA"
        except KeyError:
            continue
        
        transformed.append(value)
    
    return '\t'.join(map(str, transformed))



""" unit test the  class"""

import os
import unittest
import asyncio
import math

from denovonear.rate_limiter import RateLimiter
from denovonear.cluster_test import fishers_method, cluster_de_novos

async def call(func, *args, **kwargs):
    ''' call ensembl rest API function
    '''
    async with RateLimiter(15) as ensembl:
        return await func(*args, ensembl=ensembl, **kwargs)

def _run(func, *args, **kwargs):
    return asyncio.get_event_loop().run_until_complete(call(func, *args, **kwargs))

class TestClusterTestPy(unittest.TestCase):
    """ unit test the genic test
    """
    
    def test_cluster_de_novos(self):
        """ check that cluster_de_novos() runs correctly
        """
        
        symbol = 'PPP2R5D'
        de_novos = {'missense': [42975003, 42975003, 42975003, 42975013],
            'nonsense': []}
        
        p_values = _run(cluster_de_novos, symbol, de_novos, iterations=1000000)
        
        self.assertAlmostEqual(p_values['miss_prob'], 3e-06, delta=3e-6)
        self.assertEqual(p_values['miss_dist'], '2.3')
        self.assertTrue(math.isnan(p_values['nons_prob']))
        self.assertEqual(p_values['nons_dist'], 'nan')
        
        # TODO: add a test with de novos that occur on different transcripts
    
    def test_fishers_method(self):
        """ check that fishers combined method calculates correctly
        """
        
        p_values = [0.01, 0.001]
        self.assertAlmostEqual(fishers_method(p_values), 0.0001251292546)
        
        # correct for single value. This is off by a minuscule fraction due to
        # float division
        p_values = [0.01]
        self.assertAlmostEqual(fishers_method(p_values), 0.01)
        
        # correct for list with NA value
        p_values = [0.01, float('nan')]
        self.assertAlmostEqual(fishers_method(p_values), 0.01)
        
        # correct for list with NA value
        p_values = [0.01, 0.001, float('nan')]
        self.assertAlmostEqual(fishers_method(p_values), 0.0001251292546)
        
        # correct without any p-values
        p_values = [float('nan'), float('nan')]
        self.assertTrue(math.isnan(fishers_method(p_values)))
        
        # raise error for list with zero
        with self.assertRaises(ValueError):
            p_values = [0, 0.01]
            fishers_method(p_values)



"""
Copyright (c) 2015 Genome Research Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from __future__ import print_function

import os
import unittest
import itertools

from denovonear.site_specific_rates import get_gene_range, get_mutated_aa, SiteRates
from denovonear.weights import WeightedChoice
from denovonear.transcript import Transcript

def generate_rates(kmer_length=3):
    ''' make a list of tuples for all sequence alterations of a given length
    
    Returns:
        a pseudo mutation rate dictionary, filled with initial sequence,
        modified sequence and mutation rate.
    '''
    
    bases = set(["A", "T", "G", "C"])
    
    rates = []
    for initial in itertools.product("".join(bases), repeat=kmer_length):
        trinuc = "".join(initial).encode('utf8')
        
        rate = '5e-7'.encode('utf8')
        for base in bases - set(initial[1]):
            changed = list(initial)
            changed[kmer_length // 2] = base
            changed = "".join(changed).encode('utf8')
            
            rates.append([trinuc, changed,  rate])
    
    return rates

class TestSiteRatesPy(unittest.TestCase):
    """ unit test the SiteRates class
    """
    
    rates = generate_rates()
    
    def setUp(self):
        """ construct a Transcript object to add sequence to
        """
        
        self.transcript = self.construct_gene()
        
        cds = "ATGTCCATAACCAAAGCCTGA"
        genomic = "CCTCCAGATTCACGGGAAGCATGTCCATAAGTAGGGAGATATTTGGTGCTCTCATTTG" \
            "TGGAGACTCTAGCCAAAGCCTGAGTCATGCGTACCATAGATAG"
        
        self.transcript.add_cds_sequence(cds)
        self.transcript.add_genomic_sequence(genomic, offset=10)
        
        self.weights = SiteRates(self.transcript, self.rates)
    
    def construct_gene(self, name='TEST', chrom='1', start=100, end=179,
            strand='+', exons=[(100, 119), (160, 179)],
            cds=[(110, 119), (160, 170)]):
        
        tx = Transcript(name, chrom, start, end, strand)
        tx.set_exons(exons, cds)
        tx.set_cds(cds)
        
        return tx
    
    def test_longer_mutation_rate_sequences(self):
        """ check that we can construct a SiteRates object if using a mutation
        rate dictionary with longer kmers (e.g 5-mers, rather than the 3-mers
        for the trinucleotide base rates.)
        """
        
        five_mers = generate_rates(5)
        seven_mers = generate_rates(7)
        
        transcript = self.construct_gene()
        
        cds = "ATGTCCATAACCAAAGCCTGA"
        genomic = "CCTCCAGATTCACGGGAAGCATGTCCATAAGTAGGGAGATATTTGGTGCTCTCATTTG" \
            "TGGAGACTCTAGCCAAAGCCTGAGTCATGCGTACCATAGATAG"
        
        transcript.add_cds_sequence(cds)
        transcript.add_genomic_sequence(genomic, offset=10)
        
        weights = SiteRates(transcript, five_mers)
        weights = SiteRates(transcript, seven_mers)
    
    def test_get_boundary_distance(self):
        """ check the function to get distances to the nearest intron/exon boundary
        """
        
        # check a site upstream of the gene
        self.assertEqual(self.transcript.get_boundary_distance(50), 50)
        
        # check a site at the start of a gene
        self.assertEqual(self.transcript.get_boundary_distance(100), 0)
        
        # check some sites within the first exon
        self.assertEqual(self.transcript.get_boundary_distance(110), 10)
        self.assertEqual(self.transcript.get_boundary_distance(115), 5)
        
        # check sites in the first intron
        self.assertEqual(self.transcript.get_boundary_distance(125), 6)
        self.assertEqual(self.transcript.get_boundary_distance(140), 20)
        
        # check a site in the first exon, as it becomes closer to the next intron
        self.assertEqual(self.transcript.get_boundary_distance(141), 19)
        
        # check a site downstream of the gene
        self.assertEqual(self.transcript.get_boundary_distance(200), 21)
    
    def test_get_codon_info(self):
        """ check the function that checks the codon information for a position
        """
        
        # make sure a site well outside the gene raises an error
        with self.assertRaises(ValueError):
            self.transcript.get_codon_info(50)
        
        # a position near the start site, but upstream of the CDS will raise a
        # different error
        with self.assertRaises(ValueError):
            self.transcript.get_codon_info(95)
        
        # check the first base of the CDS
        self.assertEqual(self.transcript.get_codon_info(110),
            {'cds_pos': 0, 'codon_seq': 'ATG', 'intra_codon': 0,
                "codon_number": 0, 'initial_aa': 'M', 'offset': 0})
        
        # check the second base of the CDS
        self.assertEqual(self.transcript.get_codon_info(111),
            {'cds_pos': 1, 'codon_seq': 'ATG', 'intra_codon': 1,
                "codon_number": 0, 'initial_aa': 'M', 'offset': 0})
        
        # check the third base of the CDS
        self.assertEqual(self.transcript.get_codon_info(112),
            {'cds_pos': 2, 'codon_seq': 'ATG', 'intra_codon': 2,
                "codon_number": 0, 'initial_aa': 'M', 'offset': 0})
        
        # check the fourth base of the CDS
        self.assertEqual(self.transcript.get_codon_info(113),
            {'cds_pos': 3, 'codon_seq': 'TCC', 'intra_codon': 0,
                "codon_number": 1, 'initial_aa': 'S', 'offset': 0})
        
        # check a site 2 bp into the first intron. We assign this as the
        # position of the closest exon boundary, but without any codon info
        self.assertEqual(self.transcript.get_codon_info(122),
            {'cds_pos': 9, 'codon_seq': None, 'intra_codon': None,
                "codon_number": None, 'initial_aa': None, 'offset': 3})
    
    def test_site_rates_weights(self):
        """ check the cumulative mutation rates for each consequence type.
        """
        
        wts = self.weights
        
        # these cumulative mutation rates should be correct for each consequence
        # type
        self.assertAlmostEqual(wts["missense"].get_summed_rate(), 2.45e-05, places=7)
        self.assertAlmostEqual(wts["nonsense"].get_summed_rate(), 5e-07, places=7)
        self.assertAlmostEqual(wts["synonymous"].get_summed_rate(), 4e-06, places=7)
        self.assertAlmostEqual(wts["loss_of_function"].get_summed_rate(), 6.5e-06, places=7)
        self.assertAlmostEqual(wts["splice_lof"].get_summed_rate(), 6e-06, places=7)
        self.assertAlmostEqual(wts["splice_region"].get_summed_rate(), 2.05e-05, places=7)
    
    def test_site_rates_sampled(self):
        """ check the sites sampled for each consequence group.
        
        Repeatedly sample sites within the transcript. Given enough samples, we
        should saturate the transcript, and so we can check that we have sampled
        all of those sites.
        """
        
        wts = self.weights
        n = 10000
        
        # there are numerous sites for mutating to a missense change
        self.assertEqual(set([ wts["missense"].choice() for x in range(n) ]),
            set([0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16, 18, 19, 20]))
        
        # the transcript only has one position where we can get a stop_gained
        self.assertEqual(set([ wts["nonsense"].choice() for x in range(n) ]),
            set([12]))
        
        # the transcript only has a few positions where we can get synonymous changes
        self.assertEqual(set([ wts["synonymous"].choice() for x in range(n) ]),
            set([5, 14, 17, 19]))
        
        # the transcript only has one splice donor site and one splice acceptor
        # site. The splice lof sites are shifted to the nearest exon coordinate
        # in order to be able to check CDS proximity.
        self.assertEqual(set([ wts["splice_lof"].choice() for x in range(n) ]),
            set([9, 10]))
        
        # loss-of-function sites are the union of nonsense and splice_lof sites
        self.assertEqual(set([ wts["loss_of_function"].choice() for x in range(n) ]),
            set([9, 10, 12]))
        
        # splice region variants can occur at the two bp inside the exon, or
        # within 8 bp of the intron/exon boundary (but beyond the splice lof
        # positions). The intronic positions are shifted to the nearest exon
        # coordinate for CDS proximity checking.
        self.assertEqual(set([ wts["splice_region"].choice() for x in range(n) ]),
            set([8, 9, 10, 11]))
    
    def test_site_rates_sampled_genomic_coords(self):
        ''' check the sites sampled when we request genomic coordinates.
        '''
        wts = SiteRates(self.transcript, self.rates, cds_coords=False)
        n = 10000
        
        self.assertEqual(set([ wts["missense"].choice() for x in range(n) ]),
            set([110, 111, 112, 113, 114, 116, 117, 118, 119, 160, 162, 163,
            164, 165, 166, 168, 169, 170]))
        
        # the transcript only has one position where we can get a stop_gained
        self.assertEqual(set([ wts["nonsense"].choice() for x in range(n) ]),
            set([162]))
    
    def test_get_mutated_aa(self):
        """ check that mutating a codon gives the expected amino acids
        """
        
        # check some codon mutations, including a stop mutation
        self.assertEqual(get_mutated_aa(self.transcript, "C", "AAA", 2), "N")
        self.assertEqual(get_mutated_aa(self.transcript, "A", "TGG", 2), "*")
        
        # a codon mutated to itself gives the expected amino acid
        self.assertEqual(get_mutated_aa(self.transcript, "A", "AAA", 2), "K")
        
        # non-DNA codons raise errors
        with self.assertRaises(ValueError):
            get_mutated_aa(self.transcript, "C", "RRR", 2)
    
    def test_splice_lof_check(self):
        """ check that splice_lof_check() works correctly
        """
        
        # check a site within the intron, but only 2 bp away from the
        # intron/exon boundary
        self.weights.check_position(121)
        self.assertEqual(self.weights.check_consequence('', '', 121), 'splice_lof')
        
        # check a site within the exon, and also 2 bp away from the intron/exon
        # boundary
        self.weights.check_position(117)
        self.assertNotEqual(self.weights.check_consequence('', '', 117), 'splice_lof')
        
        # check a intron site outside the splice lof positions
        self.weights.check_position(122)
        self.assertNotEqual(self.weights.check_consequence('', '', 122), 'splice_lof')
    
    def test_nonsense_check(self):
        """ check that nonsense_check() works correctly
        """
        
        self.weights.check_position(161)
        self.assertEqual(self.weights.check_consequence("N", "*", 161), 'nonsense')
        self.assertNotEqual(self.weights.check_consequence("N", "G", 161), 'nonsense')
        self.assertNotEqual(self.weights.check_consequence("*", "G", 161), 'nonsense')
        self.assertNotEqual(self.weights.check_consequence("*", "*", 161), 'nonsense')
    
    def test_missense_check(self):
        """ check that missense_check() works correctly
        """
        
        # missense mutations can either be where the amino acids differ, or a
        # stop site changes to coding an amino acid (technically these are
        # stop_lost, but they carry a missense-like severity).
        self.weights.check_position(161)
        self.assertTrue(self.weights.check_consequence("N", "G", 161), 'missense')
        self.assertTrue(self.weights.check_consequence("*", "G", 161), 'missense')
        
        # don't include stop gained mutations, or stop to stop
        self.assertNotEqual(self.weights.check_consequence("N", "*", 161), 'missense')
        self.assertNotEqual(self.weights.check_consequence("*", "*", 161), 'missense')
        
        # the case below shouldn't occur, where the site is in the intron and
        # within the splice lof positions, but somehow the initial and modified
        # amino acids differ, but check this anyway.
        # self.weights.boundary_dist = 2
        self.weights.check_position(121)
        self.assertNotEqual(self.weights.check_consequence("N", "G", 121), 'missense')
    
    def test_splice_region_check(self):
        """ check that splice_region_check() works correctly
        """
        
        self.weights.check_position(123)
        self.assertEqual(self.weights.check_consequence("", "", 123), 'splice_region')
        
        # check a site in the splice lof positions
        self.weights.check_position(121)
        self.assertNotEqual(self.weights.check_consequence("", "", 121), 'splice_region')
        
        # check a site just inside the splice region positions
        self.weights.check_position(127)
        self.assertEqual(self.weights.check_consequence("", "", 128), 'splice_region')
        
        # check a site just outside the splice region positions
        self.weights.check_position(130)
        self.assertNotEqual(self.weights.check_consequence("", "", 130), 'splice_region')
        
        # check an exonic splice region position
        self.weights.check_position(117)
        self.assertEqual(self.weights.check_consequence("N", "N", 117), 'splice_region')
        
        # check an exonic site just beyond the splice region positions
        self.weights.check_position(116)
        self.assertNotEqual(self.weights.check_consequence("N", "N", 116), 'splice_region')
        
        # check an exonic site inside the splice region positions, but with
        # mutated amino acids aren't splice region variants
        self.weights.check_position(117)
        self.assertNotEqual(self.weights.check_consequence("N", "K", 117), 'splice_region')
    
    def test_synonymous_check(self):
        """ check that synonymous_check() works correctly
        """
        
        self.weights.check_position(115)
        self.assertEqual(self.weights.check_consequence("N", "N", 115), 'synonymous')
        self.assertEqual(self.weights.check_consequence("*", "*", 115), 'synonymous')
        
        # amino acid changes aren't synonymous
        self.assertNotEqual(self.weights.check_consequence("N", "*", 115), 'synonymous')
        self.assertNotEqual(self.weights.check_consequence("*", "N", 115), 'synonymous')
        
        # sites in splice region or splice lof aren't synonymous
        self.weights.check_position(117)
        self.assertNotEqual(self.weights.check_consequence("N", "N", 117), 'synonymous')
        self.weights.check_position(121)
        self.assertNotEqual(self.weights.check_consequence("", "", 121), 'synonymous')
    
    def test_get_gene_range(self):
        """ check that get_gene_range() works correctly
        """
        
        self.assertEqual(get_gene_range(self.transcript), {'start': 110, 'end': 170})
    
    def test_check_position_missense_only(self):
        """ check that check_position() works correctly for missense changes
        """
        
        self.weights.clear()
        self.weights.check_position(110)
        
        # mutating the first base of the start codon all has three missense
        # possibilities, and if all those have a uniform rate of 5e-7, then the
        # missense rate should sum to 1.5e-7
        self.assertEqual(self.weights["missense"].get_summed_rate(), 1.5e-6)
        self.assertEqual(self.weights["nonsense"].get_summed_rate(), 0)
        self.assertEqual(self.weights["synonymous"].get_summed_rate(), 0)
        self.assertEqual(self.weights["splice_lof"].get_summed_rate(), 0)
        self.assertEqual(self.weights["splice_region"].get_summed_rate(), 0)
        self.assertEqual(self.weights["loss_of_function"].get_summed_rate(), 0)
        
    def test_check_position_mixed_nonsense(self):
        """ check that check_position() works correctly for mixed changes
        """
        
        self.weights.clear()
        self.weights.check_position(162)
        
        # check a position where one alternate base gives a nonsense change
        self.assertEqual(self.weights["missense"].get_summed_rate(), 1.0e-6)
        self.assertEqual(self.weights["nonsense"].get_summed_rate(), 0.5e-6)
        self.assertEqual(self.weights["synonymous"].get_summed_rate(), 0)
        self.assertEqual(self.weights["splice_lof"].get_summed_rate(), 0)
        self.assertEqual(self.weights["splice_region"].get_summed_rate(), 0)
        self.assertEqual(self.weights["loss_of_function"].get_summed_rate(), 0.5e-6)
    
    def test_check_position_noncoding(self):
        """ check that check_position() works correctly for noncoding changes
        """
        
        self.weights.clear()
        categories = ['missense', 'nonsense', 'synonymous', 'splice_lof', 'splice_region']
        
        # a deep intronic site won't alter the summed rates
        self.weights.check_position(140)
        for x in categories:
            self.assertEqual(self.weights[x].get_summed_rate(), 0)
        
        # an upstream site won't alter the summed rates
        self.weights.check_position(self.transcript.get_start() - 1)
        for x in categories:
            self.assertEqual(self.weights[x].get_summed_rate(), 0)
        
        # a downstream site won't alter the summed rates
        self.weights.check_position(self.transcript.get_end() + 1)
        for x in categories:
            self.assertEqual(self.weights[x].get_summed_rate(), 0)
        
        # a splice lof only affects certain rates
        self.weights.check_position(121)
        self.assertEqual(self.weights["missense"].get_summed_rate(), 0)
        self.assertEqual(self.weights["nonsense"].get_summed_rate(), 0)
        self.assertEqual(self.weights["synonymous"].get_summed_rate(), 0)
        self.assertEqual(self.weights["splice_lof"].get_summed_rate(), 1.5e-6)
        self.assertEqual(self.weights["splice_region"].get_summed_rate(), 0)
        self.assertEqual(self.weights["loss_of_function"].get_summed_rate(), 1.5e-6)
        
        # a splice region only affects certain rates
        self.weights.clear()
        self.weights.check_position(125)
        self.assertEqual(self.weights["missense"].get_summed_rate(), 0)
        self.assertEqual(self.weights["nonsense"].get_summed_rate(), 0)
        self.assertEqual(self.weights["synonymous"].get_summed_rate(), 0)
        self.assertEqual(self.weights["splice_lof"].get_summed_rate(), 0)
        self.assertEqual(self.weights["splice_region"].get_summed_rate(), 1.5e-6)
        self.assertEqual(self.weights["loss_of_function"].get_summed_rate(), 0)



""" Obtains genomic sequence, transcript IDs, gene, exon and CDS coordinates for
transcripts from Ensembl.
"""

import json
import asyncio

def get_base_url(build):
    assert build in ["grch37", "grch38"], f'unknown build: {build}'
    ver = build + '.' if build == "grch37" else ''
    return f'https://{ver}rest.ensembl.org'

async def get_genes_for_hgnc_id(ensembl, hgnc_symbol, build='grch37'):
    """ obtain the ensembl gene IDs that correspond to a HGNC symbol
    
    e.g. http://rest.ensembl.org/xrefs/symbol/homo_sapiens/KMT2A
    """
    headers = {"content-type": "application/json"}
    ext = f"xrefs/symbol/homo_sapiens/{hgnc_symbol}"
    url = f'{get_base_url(build)}/{ext}'
    resp = await ensembl.get(url, headers)
    return [x['id'] for x in json.loads(resp) if x['type'] == 'gene']

async def get_transcript_ids_for_ensembl_gene_id(ensembl, gene_id, symbols, build='grch37'):
    """ fetch the ensembl transcript IDs for a given ensembl gene ID.
    
    Args:
        gene_id: Ensembl gene ID
        symbols: HGNC symbol(s) for gene. We can have multiple symbols if we
            pull previous symbols out.
    """
    chroms = {"1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", \
         "12", "13", "14", "15", "16", "17", "18", "19", "20", "21", "22", \
          "X", "Y"}
    headers = {"content-type": "application/json"}
    
    transcript_ids = []
    ext = f"overlap/id/{gene_id}?feature=transcript"
    url = f'{get_base_url(build)}/{ext}'
    resp = await ensembl.get(url, headers)
    
    for item in json.loads(resp):
        # ignore non-coding transcripts
        if item["biotype"] not in ["protein_coding", "polymorphic_pseudogene"]:
            continue
        
        # ignore transcripts not on the standard chromosomes
        if item["Parent"] != gene_id or item["seq_region_name"] not in chroms or \
                all([s not in item["external_name"] for s in symbols]):
            continue
        transcript_ids.append(item["id"])
    
    return transcript_ids

async def get_previous_symbol(ensembl, hgnc_symbol):
    """ sometimes we get HGNC symbols that do not match the ensembl rest version
    that we are currently using. We can look for earlier HGNC symbols for
    the gene using the service at rest.genenames.org
    
    Args:
        hgnc_symbol: HGNC symbol for the gene (eg "MLL2")
    
    Returns:
        list of deprecated gene symbols (eg ["KMT2A"])
    """
    headers = {"accept": "application/json"}
    ext = f"fetch/symbol/{hgnc_symbol}"
    url = f'https://rest.genenames.org/{ext}'
    resp = await ensembl.get(url, headers)
    data = json.loads(resp)
    
    docs = data["response"]["docs"]
    # strip out any gene entries that have been invalidated
    docs = [ x for x in docs if x["status"] != "Entry Withdrawn"]
    
    if len(docs) == 0:
        return []
    elif len(docs) > 1:
        raise ValueError(f"{hgnc_symbol} has more than one alternate symbol.")
    elif "prev_symbol" in docs[0]:
        return docs[0]["prev_symbol"]
    return []

async def get_genomic_seq_for_transcript(ensembl, transcript_id, expand, build='grch37'):
    """ obtain the sequence for a transcript from ensembl
    """
    headers = {"content-type": "application/json"}
    ext = f"sequence/id/{transcript_id}?type=genomic;expand_3prime={expand};expand_5prime={expand}"
    url = f'{get_base_url(build)}/{ext}'
    resp = await ensembl.get(url, headers)
    
    gene = json.loads(resp)
    if gene["id"] != transcript_id:
        raise ValueError("ensembl gave the wrong transcript")
    
    _, _, chrom, start, end, strand = gene["desc"].split(":")
    start = int(start) + expand
    end = int(end) - expand
    strand = "+" if strand != '-1' else "-"
    
    return (chrom, start, end, strand, gene["seq"])

async def get_cds_seq_for_transcript(ensembl, transcript_id, build='grch37'):
    """ obtain the sequence for a transcript from ensembl
    """
    headers = {"content-type": "text/plain"}
    ext = f"sequence/id/{transcript_id}?type=cds"
    url = f'{get_base_url(build)}/{ext}'
    resp = await ensembl.get(url, headers)
    return resp.decode('utf8')

async def get_protein_seq_for_transcript(ensembl, transcript_id, build='grch37'):
    """ obtain the sequence for a transcript from ensembl
    """
    headers = {"content-type": "text/plain"}
    ext = f"sequence/id/{transcript_id}?type=protein"
    url = f'{get_base_url(build)}/{ext}'
    resp = await ensembl.get(url, headers)
    return resp.decode('utf8')

async def get_ranges_for_tx(ensembl, transcript_id, feature, build='grch37'):
    """ get coordinates for exons (just cds or full exonic)
    """
    headers = {"content-type": "application/json"}
    ext = f"overlap/id/{transcript_id}?feature={feature}"
    url = f'{get_base_url(build)}/{ext}'
    resp = await ensembl.get(url, headers)
    
    ranges = []
    for exon in json.loads(resp):
        if exon["Parent"] != transcript_id:
            continue
        ranges.append((exon["start"], exon["end"]))
    
    return ranges

async def get_exon_ranges_for_transcript(ensembl, transcript_id, build='grch37'):
    """ obtain the sequence for a transcript from ensembl
    """
    return await get_ranges_for_tx(ensembl, transcript_id, 'exon', build)

async def get_cds_ranges_for_transcript(ensembl, transcript_id, build='grch37'):
    """ obtain the sequence for a transcript from ensembl
    """
    return await get_ranges_for_tx(ensembl, transcript_id, 'cds', build)



""" unit test the WeightedChoice class"""

from __future__ import division

import os
import unittest
import math

from denovonear.weights import WeightedChoice
from denovonear.transcript import Transcript
from denovonear.site_specific_rates import SiteRates
from denovonear.load_mutation_rates import load_mutation_rates
from denovonear.simulate import get_p_value

class TestGetPValuePy(unittest.TestCase):
    """ unit test the simulation of p-values
    """
    
    def setUp(self):
        
        self.transcript = self.construct_gene()
        self.rates = self.get_rates(self.transcript)
    
    def construct_gene(self):
        
        chrom = "1"
        name = "TEST"
        strand = "+"
        start = 0
        end = 70
        exons = [(5, 58)]
        cds = [(5, 58)]
        transcript = Transcript(name, chrom, start, end, strand)
        transcript.set_exons(exons, cds)
        transcript.set_cds(cds)
        
        cds = "ATGTGGGCTCCACCAGCAGCAATCATGGGATGGGCCCACCAAGAAGGTGGGTAA"
        gdna = "GGGGGATGTGGGCTCCACCAGCAGCAATCATGGGATGGGCCCACCAAGAAGGTGGGTAACCAGGCCCC"
        transcript.add_cds_sequence(cds)
        transcript.add_genomic_sequence(gdna)
        
        return transcript
    
    def get_rates(self, tx):
        # load the sequence contect mutation rates, then assess each site in the
        # CDS.
        mut_dict = load_mutation_rates()
        
        return SiteRates(tx, mut_dict)
    
    def test_get_p_value_single_de_novo(self):
        """ check that get_p_value() works correctly
        """
        
        # check that we don't assess transcripts with a single de novo
        iterations = 10000
        cq = 'missense'
        de_novos = [5]
        
        (obs, p_value) = get_p_value(self.transcript, self.rates, iterations, cq, de_novos)
        self.assertTrue(math.isnan(obs))
        self.assertTrue(math.isnan(p_value))
    
    def test_get_p_value(self):
        """
        """
        
        iterations = 10000
        cq = 'missense'
        de_novos = [5, 5]
        (obs, p_value) = get_p_value(self.transcript, self.rates, iterations, cq, de_novos)
        
        self.assertTrue(p_value < 0.04)
        self.assertEqual(obs, '0.0')
    
    def test_get_p_value_nonsignificant(self):
        """ check for de novos spread across the gene
        """
        
        iterations = 10000
        cq = 'missense'
        de_novos = [5, 58]
        
        (obs, p_value) = get_p_value(self.transcript, self.rates, iterations, cq, de_novos)
        
        self.assertTrue(p_value == 1.0)
        self.assertEqual(obs, '53.0')
    
    def test_get_p_value_lofs(self):
        """ check for loss of function rates
        """
    
        # construct two different tx objects, for the same gene, since otherwise
        # the p-values from different runs are the same
        tx1 = self.construct_gene()
        rates1 = self.get_rates(tx1)
        
        iterations = 10000
        cq = 'loss_of_function'
        de_novos = [5, 6]
        (obs_1, p_1) = get_p_value(tx1, rates1, iterations, cq, de_novos)
        
        tx2 = self.construct_gene()
        rates2 = self.get_rates(tx2)
        
        # make sure we can use the string 'lof' to get loss-of-function rates.
        # NOTE: due to randomly sampling, this will fail ~0.1% of the time,
        # purely by chance. If this fails, first try rerunning the tests.
        iterations = 10000
        cq = 'lof'
        (obs_2, p_2) = get_p_value(tx2, rates2, iterations, cq, de_novos)
        self.assertEqual(obs_1, obs_2)
        self.assertTrue(abs(p_1 - p_2) < 0.017)
